#----------------------------------------------------------------
# Copyright (c) 2002 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------

=head2 ExtraGUI.pm

This module contains a bunch of miscellaneous GUI code for making
windows other than the main Browser.pm window, e.g. dialog boxes.
For the most part, Browser.pm calls routines in ExtraGUI.pm,
not the other way around.
The only exceptions are (1) we do calls to
 Browser::empty_toplevel_window(); (2) we make use
of $Browser::words.

=cut

use strict;

use Tk;
use Browser;
use Words;
use MyWords;
use Input;
use FileDialogPatched;
use Preferences;

package ExtraGUI;

# See the Words and MyWords modules for info on the following.
our $words_prefix;		# initialized in each routine
sub w {
    my $key = shift;
    my $newline = "";
    if ($key =~ m/\n$/) {$newline="\n"}
    chomp $key;
    return $Browser::words->get($words_prefix.".".$key).$newline;
}


BEGIN {
  my %fonts = ();
  my %size    = ('plain' => 11,          'bold' => 11,          'small_plain' => 10,         'fixed_width'=>11);
  my %family  = ('plain' => 'helvetica', 'bold' => 'helvetica', 'small_plain' => 'helvetica','fixed_width'=>'courier');
  my %weight  = ('plain' => 'normal',    'bold' => 'bold',      'small_plain' => 'normal',   'fixed_width'=>'normal');
  my %slant   = ('plain' => 'roman',     'bold' => 'roman',     'small_plain' => 'roman',    'fixed_width'=>'roman');
  my $widget; # Gets initialized by calling fonts_init. We just need to have some widget
              # that we can call the fontCreate() method on.
  sub fonts_init {
    $widget = shift;
  }
  sub font {
    my $style = shift;
    if (!exists($fonts{$style})) {
      $fonts{$style} = $widget->fontCreate(-size=>$size{$style},-family=>$family{$style},-weight=>$weight{$style},-slant=>$slant{$style});
    }
    return $fonts{$style};
  }
}

=head3 preferred_location

Returns a geometry string describing where dialog boxes should be placed: in front of the main window,
and just a little lower at the top, like ``sheets'' in MacOS X.

=cut

sub preferred_location {
  my ($x,$y) = preferred_location_coords();
  # For some reason, the main window's geometry is returned as the coordinates of the /interior/
  # of the window. Therefore, we don't really need to add anything to $y.
  return coords_to_geometry_string($x,$y);
}

sub offset_location {
  my $xoffset = shift; my $yoffset = shift;
  my ($x,$y) = preferred_location_coords();
  return coords_to_geometry_string($x+$xoffset,$y+$yoffset);
}

sub coords_to_geometry_string {
  my $x = shift; my $y = shift;
  if (! ($y =~ m/^[\+\-]/)) {$y='+'.$y}
  if (! ($x =~ m/^[\+\-]/)) {$x='+'.$x}
  return $x.$y;
}

sub preferred_location_coords {
  Browser::main_window()->geometry() =~ m/^=?(\d+)x(\d+)([+-]\d+)([+-]\d+)$/;
  return ($3,$4);
}

=head3 choose_file()

Displays a file dialog box.

=cut

sub choose_file {
  my %args = (
    CALLBACK=>sub{},
    TITLE=>'',
    CREATE=>0,
    PATH=>'.',
    @_,
  );
  my $callback = $args{CALLBACK};
  my $path = $args{PATH};
  #print "path=$path\n";
  my $dlog = Browser::main_window()->FileDialog(-Title =>$args{TITLE}, -Create => $args{CREATE});
  $dlog->configure(-FPat => '*.gb',-ShowAll => 'NO',-Path=>$path);
  my $file_name = $dlog->Show();
  $dlog->geometry(preferred_location());
  $file_name =~ s|^\.\/||; # Strip a leading ./
  if (!$file_name) {$file_name = ""}; # If they hit cancel, we get undef.
  if ($file_name ne "" && !($file_name =~ m/\.gb$/)) {$file_name = $file_name . ".gb"}
  &$callback($file_name);
}


=head3 error_message()

Takes one parameter, which is the error message.

=cut

sub error_message {
  my $message = shift;
  local $words_prefix = "b.error_message";
  my $box = Browser::empty_toplevel_window(w('error'));
  $box->Label(-text=>$message,-justify=>'left',-font=>font('plain'))->pack();
  $box->Button(-text=>w("ok"),command=>sub{$box->destroy()})->pack();
  $box->geometry(preferred_location());
  beep_if_allowed($box);
}


=head3 message()

Takes one parameter, which is the message. Cf. error_message().


=cut

sub message {
  my $message = shift;
  local $words_prefix = "b.error_message";
  my $box = Browser::empty_toplevel_window('');
  $box->geometry(preferred_location());
  $box->Label(-text=>$message,-justify=>'left',-font=>font('plain'))->pack();
  $box->Button(-text=>w("ok"),command=>sub{$box->destroy()})->pack();
}

=head3 confirm()

First argument is a question, e.g., "Drop Al Einstein?"
Second argument is a callback, which gets
1 if they hit ok, 0 if they cancel.
This can also be used for other kinds of two-choice dialog boxes:
the third and fourth arguments, if present, are text to use
instead of OK and Cancel.

=cut

sub confirm {
  my $question = shift;
  my $callback = shift;
  local $words_prefix = "b.dialog";
  my $ok_text = w('ok');
  if (@_) {$ok_text = shift}
  my $cancel_text = w('cancel');
  if (@_) {$cancel_text = shift}
  my $box = Browser::empty_toplevel_window(w('confirm'));
  $box->geometry(preferred_location());
  $box->Label(-text=>$question,-justify=>'left',-font=>font('plain'))->pack();
  $box->Button(-text=>$ok_text,command=>sub{$box->destroy();&$callback(1)})->pack();
  $box->Button(-text=>$cancel_text,command=>sub{$box->destroy();&$callback(0)})->pack();
}

=head3 ask()

Get a string as input in a dialog box.

    PROMPT=>"?",
    TITLE=>"",
    WIDTH=>40,
    CALLBACK=>sub{},
    ACTION_BUTTON=>w("ok"),
    DEFAULT=>'',
    PASSWORD=>0

The callback routine gets the result, or gets no argument if the
user hit cancel.

=cut

sub ask {
  local $words_prefix = "b.dialog";
  my %args = (
    PROMPT=>"?",
    TITLE=>"",
    WIDTH=>40,
    CALLBACK=>sub{},
    ACTION_BUTTON=>w("ok"),
    DEFAULT=>'',
    PASSWORD=>0,
    BEEP=>0, # If set to 1, /and/ if user's prefs allow it, beep.
    @_
  );
  my $callback = $args{CALLBACK};
  my $result = $args{DEFAULT};
  my $pwd = $args{PASSWORD};
  my $beep = $args{BEEP};
  my @pwd_option = ();
  if ($pwd) {@pwd_option=(-show=>'*')}
  my $box = Browser::empty_toplevel_window($args{TITLE});
  $box->geometry(preferred_location());
  $box->Label(-anchor=>'w',-justify=>'left',-text=>$args{PROMPT})->pack();
  my $f = $box->Frame->pack(-side => 'right');
  my $e = $box->Entry(
	      -takefocus=>1,
	      -width=>$args{WIDTH},
	      -textvariable=>\$result,
              @pwd_option,
  )->pack(-side=>'left');
  $e->focus();
  $e->icursor('end');
  $f->Button(-text=>w("cancel"),command=>sub{$box->destroy();&$callback()})->pack(-side=>'right');
  $e->bind('<KeyRelease-Return>'=>sub{$box->destroy();&$callback($result)});
  $f->Button(-text=>w("ok"),command=>sub{$box->destroy();&$callback($result)})->pack(-side=>'right');
  if ($beep) {beep_if_allowed($box)}
}

sub save_plain_text_to_file {
  my $text = shift;
  my $path = '.';
  if (@_) {$path = shift}
  local $words_prefix = "b.dialog";
  my $filename = $path;
  if ($path eq '.') {
    $filename = '';
  }
  if ($filename ne '' && !($filename =~ m@/$@)) {$filename="$filename/"}
  my $box = Browser::empty_toplevel_window(w('save_as'));
  $box->geometry(preferred_location());
  my $entry = $box->Entry(-textvariable=>\$filename,-width=>100);
  $entry->pack()->focus();
  $entry->icursor('end');
  my $f = $box->Frame();
  my $do_it = sub {
              local $words_prefix = "b.dialog";
              my $really_do_it = shift;
              if ($really_do_it == 1) {
	        my $err = '';
                if ($err eq '') {
                  if (open(FILE,">$filename")) {
                    print FILE $text;
                    close(FILE);
                  }
                  else {
		    $err = sprintf w('error_opening_for_output'),$filename;
                  }
                }
                if ($err ne '') {error_message($err)}
	      }
  };
  $f->Button(-text=>w('save'),
       -command=>sub{
              local $words_prefix = "b.dialog";
	      my $err = '';
	      if ($filename =~ m/.gb$/) {$err = w('save_on_top_of_gb')}
              if ($err eq '') {
                if (-e $filename) {
                  confirm((sprintf w('confirm_overwrite'),$filename),$do_it);
                }
                else {
                  &$do_it('y')
                }
              }
              if ($err ne '') {error_message($err)}
              $box->destroy();
          }
     )->pack(-side=>'left');
  $f->Button(-text=>w('cancel'),-command=>sub{$box->destroy()})->pack(-side=>'left');
  $f->pack();
}

sub show_text {
  local $words_prefix = "b.dialog";
  my %args = (
    TITLE=>"",
    TEXT=>"",
    WIDTH=>60,
    MIN_HEIGHT=>3,
    MAX_HEIGHT=>50,
    BGCOLOR=>"white",
    PATH=>'.', # default directory for saving
    @_,
  );
  my $text = $args{TEXT};
  my $width = $args{WIDTH};
  my $nlines = 0;
  my @lines = split /\n/,$text; # in scalar context, we get the number of elements in the array
  foreach my $line(@lines) {
    $nlines = $nlines + int(length($line)/$width) + 1;
  }
  my $height = $nlines;
  if ($height>$args{MAX_HEIGHT}) {$height=$args{MAX_HEIGHT}}
  if ($height<$args{MIN_HEIGHT}) {$height=$args{MIN_HEIGHT}}
  my $box = Browser::empty_toplevel_window($args{TITLE});
  $box->geometry(preferred_location());
  my $t = $box->Scrolled("Text",
	-scrollbars=>'e',
        -width=>$width,
        -height=>$height,
	-background=>$args{BGCOLOR},
  )->pack();
  $t->insert('end',$text);
  # Note: The following is commented out because even though this routine is meant to be used
  # with immutable text, if you disable the Text object, they can't copy it to the clipboard.
  #  $t->configure(-state => 'disabled'); # so they can't type in it
  my $f = $box->Frame();
  $f->Button(-text=>w('ok'),-command=>sub{$box->destroy()})->pack(-side=>'left');
  $f->Button(-text=>w('save'),-command=>sub{save_plain_text_to_file($args{TEXT},$args{PATH}); $box->destroy()})->pack(-side=>'left');
  $f->pack();
}

=head3 fill_in_form

Arguments:

      INPUTS=>[], # ref to an array of Input objects
      TITLE=>"",
      INFO=>'',   # information displayed at the top
      CALLBACK=>sub{},
      CANCEL_CALLBACK=>sub{},
      COLUMNS=>2, # 2=prompts side by side with blanks, 1=vertical arrangement
      WIDTH=>80,

This routine creates a dialog box and asks for the inputs referred to in the INPUTS array.
Then it calls the callback routine, giving it a
reference to a hash containing the strings given by the
user. The hash keys are defined by the {KEY} instance variables of the Input objects.

=cut

sub fill_in_form {
  my %args = (
 	      INPUTS=>[], # ref to an array of Input objects
	      TITLE=>"",
              INFO=>'',   # information displayed at the top
	      CALLBACK=>sub{},
              CANCEL_CALLBACK=>sub{},
	      COLUMNS=>2, # 2=prompts side by side with blanks, 1=vertical arrangement
	      WIDTH=>80,
	      HEIGHT=>'',
              XOFFSET=>0,
              YOFFSET=>0,
	      @_,
  );
  my $callback = $args{CALLBACK};
  my $cancel_callback = $args{CANCEL_CALLBACK};
  my $inputs = $args{INPUTS};
  my $info = $args{INFO};
  my $xoffset = $args{XOFFSET};
  my $yoffset = $args{YOFFSET};
  my $n_inputs = @$inputs;
  my $height = 5+4*$n_inputs; # doesn't really matter that much, because they get a scrollbar if it's too small
  if ($args{HEIGHT} ne '') {$height=$args{HEIGHT}}
  my %results = ();
  local $words_prefix = "b.dialog";
  my $box = Browser::empty_toplevel_window($args{TITLE});
  #print offset_location($xoffset,$yoffset)."\n";
  $box->geometry(offset_location($xoffset,$yoffset));

  if ($info ne '') {
    $box->Label(-anchor=>'w',-justify=>'left',-text=>$info,-width=>$args{WIDTH})->pack(-side=>'top');
  }

  my $f = $box->Frame->pack(-side => 'bottom');
  $f->Button(-text=>w("ok"),
     -command=>
       sub{
         my $has_errors = 0;
         my $error_text = '';
         foreach my $input(@$inputs) {
           my $key = $input->{KEY};
	   my $retrieve = $input->{RETRIEVE};
	   my $value = &$retrieve;
           $results{$key} = $value;
	   my @err_stuff = $input->check($value);
	   my $is_an_error = @err_stuff;
	   $has_errors = $has_errors || $is_an_error;
           if ($is_an_error) {
             my $fmt = shift @err_stuff;
             my $this_error = (sprintf $fmt,(@err_stuff));
	     $error_text = $error_text . $key.": ".$this_error."\n";
	   }
         }
         if (!$has_errors) {
	   $box->destroy();
	   &$callback(\%results)
         }
	 else {
	   error_message($error_text);
         }
       }
  )->pack(-side=>'left'); 
  $f->Button(-text=>w("cancel"),
     -command=>[
       sub{
	 my $box = shift;
	 my $callback = shift;
	 my $results = shift;
	 $box->destroy();
	 &$cancel_callback()
       }
       ,$box,$cancel_callback
     ]
  )->pack(-side=>'left');

  my $ncols = $args{COLUMNS};
  my $t;
  if ($ncols==2) {
    $t = $box->Scrolled("Text",
	-scrollbars=>'oe',
        -width=>$args{WIDTH},
	-height=>$height,
    )->pack();
  }
  else { # one column
    $t = $box->Scrolled("Text",
	-scrollbars=>'oe',
        -width=>$args{WIDTH},
	-height=>$height,
    )->pack();

  }
  my $first_one = 1;
  foreach my $input(@$inputs) {
    my $which = $input->{KEY};
    my $w = $t->Label(-anchor=>'w',-justify=>'left',-text=>$input->{PROMPT},-width=>$args{WIDTH}/$ncols);

    $t->windowCreate('end',-window=>$w);
    if ($args{COLUMNS}==1) {$t->insert('end',"\n");}
    $results{$which} = $input->{DEFAULT};
    if ($input->{WIDGET_TYPE} eq 'entry') {
      $w = $t->Entry(-text=>$input->{PROMPT},-width=>$args{WIDTH}/$ncols,
		   -textvariable=>\$results{$which});
      $input->{RETRIEVE} = sub{return $results{$which}}
    }
    if ($input->{WIDGET_TYPE} eq 'radio_buttons') {
      $w = $t->Frame;
      my $radio_buttons = $input->{RADIO_BUTTONS};
      my $order = $input->{RADIO_BUTTON_ORDER};
      foreach my $value(@$order) {
        $w->Radiobutton(-text=>$radio_buttons->{$value},-value=>$value,-variable=>\$results{$which})
                        ->pack(-side=>'left');
      }
      $results{$which} = $input->{DEFAULT};
      $input->{RETRIEVE} = sub{return $results{$which}}
    }
    if ($input->{WIDGET_TYPE} eq 'text') {
      $w = $t->Scrolled('Text',
      #$w = $t->Text(
	-scrollbars=>'e',
        -width=>$args{WIDTH}/$ncols);
      $input->{RETRIEVE} = sub{return $w->get('1.0','end');}
    }
    $t->windowCreate('end',-window=>$w);
    if ($first_one) {$w->focus}
    $first_one = 0;
    $t->insert('end',"\n");
  }
  $t->configure(-state => 'disabled'); # so they can't type in it
}

sub beep_if_allowed {
  my $widget = shift;
  if (want_beeps()) {$widget->bell}
}

sub want_beeps {
  return Preferences->new()->get('beep');
}

1;
